@@ -42,7 +42,7 @@ inspect the operating system and manage environment variables.
     :func:`~ae.core.start_app_service` and :func:`~ae.core.request_app_permissions`.
 
 OS information
-~~~~~~~~~~~~~~
+^^^^^^^^^^^^^^
 
 * :data:`os_platform`: a string identifying the operating system (e.g., 'linux', 'win32', 'android', 'ios').
 * :data:`os_device_id`: a string with the ID/name of the device.
@@ -53,7 +53,7 @@ OS information
 * :func:`sys_env_text`: compiles a formatted text block with system environment information, useful for logging.
 
 environment variables & `.env` files
-~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
 * :func:`env_str`: retrieves the string value of an OS environment variable, with an option to automatically convert the
   variable name to the conventional format.
@@ -114,7 +114,7 @@ dynamically inspect modules, execution frames, and variables on the call stack.
 networking utilities
 --------------------
 
-* :func:`url_failure`: determines if and why a HTTP|FTP target is unavailable.
+* :func:`url_failure`: determines if and why an HTTP|FTP target is unavailable.
 * :func:`mask_url`: hides or replaces the password/token portion of a URL for safe logging.
 
 
@@ -124,18 +124,18 @@ general utilities & helpers
 a collection of miscellaneous mathematical, date/time, and other standalone helper functions.
 
 mathematical
-~~~~~~~~~~~~
+^^^^^^^^^^^^
 
 * :func:`sign`: returns the sign of a number (-1 for negative, 0 for zero, 1 for positive).
 * :func:`round_traditional`: rounds a float value using traditional rounding rules (e.g., `0.5` rounds up).
 
 date & time
-~~~~~~~~~~~
+^^^^^^^^^^^
 * :func:`utc_datetime`: Returns the current date and time as a timezone-naive `datetime` object in UTC.
 * :func:`now_str`: creates a compact, sortable timestamp string from the current UTC time.
 
 miscellaneous
-~~~~~~~~~~~~~
+^^^^^^^^^^^^^
 * :func:`dummy_function`: a null function that accepts any arguments and returns `None`.
 
 
@@ -158,7 +158,7 @@ base constants
 predefined constants for project structure, file conventions, and default settings.
 
 project & file structure
-~~~~~~~~~~~~~~~~~~~~~~~~
+^^^^^^^^^^^^^^^^^^^^^^^^
 
 * :data:`DOCS_FOLDER`: default name for a project's documentation folder ('docs').
 * :data:`TESTS_FOLDER`: default name for a project's tests folder ('tests').
@@ -176,7 +176,7 @@ project & file structure
   :mod:`ae.updater` and :mod:`aedev.project_manager`)
 
 formats & default settings
-~~~~~~~~~~~~~~~~~~~~~~~~~~
+^^^^^^^^^^^^^^^^^^^^^^^^^^
 
 * :data:`DATE_ISO`: ISO format string for dates ("%Y-%m-%d").
 * :data:`DATE_TIME_ISO`: ISO format string for :mod:`datetime.datetime` dates ("%Y-%m-%d %H:%M:%S.%f").
@@ -200,7 +200,7 @@ simplify file system interactions with wrappers and context managers.
 * :func:`in_wd`: a context manager that temporarily switches the current working directory.
 
 os.path shortcuts
-~~~~~~~~~~~~~~~~~
+^^^^^^^^^^^^^^^^^
 
 the following are direct references to functions in the :mod:`os.path` module for convenient and quicker access:
 
@@ -218,6 +218,7 @@ the following are direct references to functions in the :mod:`os.path` module fo
 * :data:`os_path_splitext`: :func:`os.path.splitext`
 """
 # pylint: disable=too-many-lines
+import base64
 import datetime
 import getpass
 import importlib.abc
@@ -240,12 +241,12 @@ from importlib.machinery import ModuleSpec
 from inspect import getinnerframes, getouterframes, getsourcefile
 from urllib.error import HTTPError, URLError
 from urllib.parse import urlparse, urlunparse
-from urllib.request import urlopen
+from urllib.request import Request, urlopen
 from types import ModuleType
 from typing import Any, Callable, Generator, Iterable, MutableMapping, Optional, Union, cast
 
 
-__version__ = '0.3.67'
+__version__ = '0.3.72'
 
 
 os_path_abspath = os.path.abspath
@@ -428,7 +429,6 @@ def deep_dict_update(data: dict, update: dict, overwrite: bool = True):
             data[upd_key] = upd_val
 
 
-URI_SEP_CHAR = '⫻'  # U+2AFB: TRIPLE SOLIDUS BINARY RELATION
 # noinspection GrazieInspection
 ASCII_UNICODE = (
     ('/', '⁄'),     # U+2044: Fraction Slash; '∕' U+2215: Division Slash; '⧸' U+29F8: Big Solidus;
@@ -466,13 +466,19 @@ ASCII_UNICODE = (
                     # ' ' U+202F: Narrow No-Break Space (NNBSP); ' ' U+205F Medium Mathematical Space;
                     # '␠' U+2420 symbol for space; '␣' U+2423 Open Box; '　' U+3000: Ideographic Space
     (chr(127), '␡'),  # U+2421: DELETE SYMBOL
-    # ('_', '𛲖'), # U+1BC96: Duployan Affix Low Line; '＿' U+FF3F Fullwidth Low Line
-)
-""" transformation table of special ASCII to Unicode alternative character,
+    # ('_', '𛲖'),     # U+1BC96: Duployan Affix Low Line; '＿' U+FF3F Fullwidth Low Line
+) + tuple((chr(low_asc_ord), chr(0x2400 + low_asc_ord)) for low_asc_ord in range(32))
+""" transformation table of special ASCII characters to a similar/alternative non-functional/-escaping Unicode char,
 see https://www.compart.com/en/unicode/category/Po and https://xahlee.info/comp/unicode_naming_slash.html (http!) """
 
-ASCII_TO_UNICODE = dict(ASCII_UNICODE)  #: map to convert ASCII to an alternative defused Unicode character
-UNICODE_TO_ASCII = {unicode_char: ascii_char for ascii_char, unicode_char in ASCII_UNICODE}     #: Unicode to ASCII map
+URI_SEP_STR = '://'             #: separator between service and address(host/path) in URIs
+URI_SEP_UNICODE_CHAR = '⫻'      #: single Unicode char for :data:`URI_SEP_STR`  U+2AFB: TRIPLE SOLIDUS BINARY RELATION
+
+ASCII_TO_UNICODE = str.maketrans(dict(ASCII_UNICODE))
+""" :func:`str.translate` map to convert ASCII to an alternative defused Unicode character - used by :func:`defuse` """
+UNICODE_TO_ASCII = str.maketrans({unicode_char: ascii_char for ascii_char, unicode_char in
+                                  ASCII_UNICODE + ((URI_SEP_STR, URI_SEP_UNICODE_CHAR), )})
+""" :func:`str.translate` Unicode to ASCII map - used by :func:`dedefuse` """
 
 
 def dedefuse(value: str) -> str:
@@ -481,15 +487,7 @@ def dedefuse(value: str) -> str:
     :param value:               string defused with the function :func:`defuse`.
     :return:                    re-activated form of the string (with all ASCII special characters recovered).
     """
-    original = ""
-    for char in value:
-        if char in UNICODE_TO_ASCII:
-            char = UNICODE_TO_ASCII[char]
-        elif 0x2400 <= (code := ord(char)) <= 0x241F:
-            char = chr(code - 0x2400)
-        original += char
-
-    return original.replace(URI_SEP_CHAR, '://')
+    return value.translate(UNICODE_TO_ASCII)
 
 
 def defuse(value: str) -> str:
@@ -515,15 +513,7 @@ def defuse(value: str) -> str:
     .. hint:: use the :func:`dedefuse` function to convert the defused string back to the corresponding URI/file-path.
 
     """
-    defused = ""
-    value = value.replace('://', URI_SEP_CHAR)  # make URIs shorter
-    for char in value:
-        if char in ASCII_TO_UNICODE:
-            char = ASCII_TO_UNICODE[char]
-        elif (code := ord(char)) <= 31:
-            char = chr(0x2400 + code)
-        defused += char
-    return defused
+    return value.replace(URI_SEP_STR, URI_SEP_UNICODE_CHAR).translate(ASCII_TO_UNICODE)  # replace makes URIs shorter
 
 
 def dummy_function(*_args, **_kwargs):
@@ -1058,8 +1048,17 @@ def parse_dotenv(file_path: str) -> dict[str, str]:
     :param file_path:           string with the name/path of an existing ``.env``/:data:`DOTENV_FILE_NAME` file.
     :return:                    dict with environment variable names and values
     """
-    env_vars: dict[str, str] = {}
+    lines = []          # unwrap multi-line .env variable values with backslash at line end (Docker/UNIX-style format)
+    prev_lines = ""
     for line in cast(str, read_file(file_path)).splitlines():
+        if line.endswith('\\'):
+            prev_lines += line[:-1]
+            continue
+        lines.append(prev_lines + line)
+        prev_lines = ""
+
+    env_vars: dict[str, str] = {}
+    for line in lines:
         match = _env_line.search(line)
         if not match:
             if not re.search(r'^\s*(?:#.*)?$', line):  # not comment or blank
@@ -1327,18 +1326,44 @@ def to_ascii(unicode_str: str) -> str:
     return "".join([c for c in nfkd_form if not unicodedata.combining(c)]).replace('ß', "ss").replace('€', "Euro")
 
 
-def url_failure(url: str, timeout: Optional[float] = None) -> str:  # pylint: disable=too-many-return-statements
+# pylint: disable-next=too-many-arguments,too-many-positional-arguments,too-many-return-statements
+def url_failure(url: str, token: str = "", username: str = "", password: str = "", git_repo: bool = False,
+                timeout: Optional[float] = None) -> str:
     """ determine if and why an FTP or HTTP[S] target is not available via a GET request.
 
-    :param url:                 URL of an target|page|file to check (not downloaded, fetching only the header).
+    :param url:                 URL of a target|page|file to check (not downloaded, fetching only the header).
+    :param token:               optional bearer token to authenticate (only for HTTPS protocol).
+    :param username:            optional username to authenticate (for HTTPS, together with the password argument).
+    :param password:            optional password to authenticate (for HTTPS, together with the username argument).
+    :param git_repo:            optimized check for Git repository HTTP servers/sites (like GitHub, GitLab, Bitbucket,
+                                Gitea, SourceHut, Mercury, etc. as long as they implement Smart HTTP).
     :param timeout:             connection timeout in seconds (see :func:`urllib.request.urlopen`).
     :return:                    empty string if target header is available, else an error description. if an
                                 FTP|HTTP response error occurred then the error/status code
                                 will be returned in the first 3 characters.
+
+    .. note::
+        credentials for server authentication can be specified either (1) embedded into the specified url argument,
+        (2) as bearer token in the token argument or (3) via the username/password arguments. in all cases the
+        functino will remove these secrets from the returned error description string.
     """
+    if git_repo:
+        if not url.endswith(".git"):
+            url += ".git"
+        url += "/info/refs?service=git-upload-pack"
+
+    headers = {}
+    if token:
+        assert not username and not password, "url_failure accepts either a token or username/password, not both"
+        headers['Authorization'] = "Bearer " + token
+    elif username or password:
+        creds = f"{username}:{password}".encode('utf-8')
+        headers['Authorization'] = "Basic " + base64.b64encode(creds).decode('utf-8')
+
     # noinspection PyBroadException
     try:
-        with urlopen(url, timeout=timeout) as response:             # open connection and read header
+        request = Request(url, method='GET', headers=headers)
+        with urlopen(request, timeout=timeout) as response:         # open connection and only read the header
             status = response.getcode()                             # no need to call response.read()
             return "" if 200 <= status < 300 else f"{status} {mask_url(url)} {response.reason=}"
 
@@ -1346,17 +1371,20 @@ def url_failure(url: str, timeout: Optional[float] = None) -> str:  # pylint: di
         return f"{exception.code} {mask_url(url)} raised HTTPError {exception.reason=}"
 
     except URLError as exception:
-        err_prefix = f"996 {mask_url(url)} raised {exception.errno=} {exception.reason=};"
+        err_msg = f" {mask_url(url)} raised {exception.errno=} {exception.reason=};"
         if isinstance(exception.reason, socket.gaierror):
-            return f"{err_prefix} could not resolve hostname"
-        if isinstance(exception.reason, socket.timeout):
-            return f"{err_prefix} connection timed out after {timeout} seconds"
+            return '995' + f"{err_msg} could not resolve hostname"
         if isinstance(exception.reason, ssl.SSLCertVerificationError):
-            return f"{err_prefix} SSL certificate verification failed"
-        return f"{err_prefix} could not reach the server"
+            return '996' + f"{err_msg} SSL certificate verification failed"
+        if isinstance(exception.reason, socket.timeout):
+            return '997' + f"{err_msg} connection timed out after {timeout} seconds"
+        return '998' + f"{err_msg} could not reach the server"
 
-    except Exception:                                               # pylint: disable=broad-exception-caught
-        return f"999 {mask_url(url)} raised unexpected exception"   # NOT put str(_exception) because contains password
+    except socket.timeout as _exception:    # noqa: F841 # str(_exception) could contain password|token
+        return '997' + f" {mask_url(url)} raised socket-timeout exception after {timeout} seconds"
+
+    except Exception as _exception:         # noqa: F841 # pylint: disable=broad-exception-caught
+        return '999' + f" {mask_url(url)} raised unexpected exception"   # str(_exception) COULD contain password
 
 
 def utc_datetime() -> datetime.datetime:
