@@ -1,7 +1,6 @@
 """ ae.base unit tests """
 import datetime
 import os
-import pytest
 import shutil
 import socket
 import ssl
@@ -9,20 +8,24 @@ import string
 import sys
 import tempfile
 import textwrap
+import time
+import timeit
 
 from collections import OrderedDict
 from configparser import ConfigParser
 # noinspection PyProtectedMember
 from http.client import HTTPMessage
 from types import ModuleType
-from typing import cast, Any
+from typing import cast, Any, Optional
 from unittest.mock import patch
 from urllib.error import HTTPError, URLError
 
+import pytest
+
 # noinspection PyProtectedMember
 from ae.base import (
-    ASCII_TO_UNICODE, BUILD_CONFIG_FILE, DOTENV_FILE_NAME, PY_EXT, PY_INIT, PY_MAIN, TESTS_FOLDER, UNICODE_TO_ASCII,
-    UNSET, URI_SEP_CHAR,
+    ASCII_TO_UNICODE, ASCII_UNICODE, BUILD_CONFIG_FILE, DOTENV_FILE_NAME, PY_EXT, PY_INIT, PY_MAIN, TESTS_FOLDER,
+    UNICODE_TO_ASCII, UNSET, URI_SEP_STR, URI_SEP_UNICODE_CHAR,
     app_name_guess, ascii_str, build_config_variable_values, camel_to_snake,
     dedefuse, deep_dict_update, defuse, dummy_function, duplicates, env_str, evaluate_literal,
     force_encoding, format_given, full_stack_trace, import_module, instantiate_config_parser, in_wd,
@@ -67,15 +70,32 @@ def os_env_test_env():
 module_test_var = 'module_test_var_val'   # used for stack_var()/try_exec() tests
 
 
-def test_unset_truthiness():
+def test_unset_truthiness_and_null_length():
     assert not UNSET
     assert bool(UNSET) is False
 
-
-def test_unset_null_length():
     assert len(UNSET) == 0
 
 
+def test_proof_os_path_shortcuts_performance_win():
+    att_call_setup = textwrap.dedent("""
+    import os
+    path1, path2, path3 = "folder1", "folder2", "file.tst"
+    """)
+    sho_call_setup = att_call_setup + textwrap.dedent("""
+    os_path_join = os.path.join
+    """)
+
+    att_call_code = "os.path.join(path1, path2, path3)"
+    sho_call_code = "os_path_join(path1, path2, path3)"
+
+    time_att = timeit.timeit(att_call_code, setup=att_call_setup, number=3_000_000)
+    time_sho = timeit.timeit(sho_call_code, setup=sho_call_setup, number=3_000_000)
+
+    assert time_sho < time_att
+    print(f"\n¡!¡!¡! os_path_* shortcuts are ~{((time_att - time_sho) / time_att) * 100:.2f}% faster")
+
+
 class TestErrorMsgMixin:
     def test_instantiation(self):
         ins = ErrorMsgMixin()
@@ -157,11 +177,11 @@ class TestBaseHelpers:
         assert all(ord(_) >= 128 for _ in uni_str)
         assert all(ord(_) < 128 for _ in ascii_str(uni_str))
 
-        uni_str = "".join(UNICODE_TO_ASCII.keys())
+        uni_str = "".join(_uco for _asc, _uco in ASCII_UNICODE)
         assert any(ord(_) >= 128 for _ in uni_str)
         assert all(ord(_) < 128 for _ in ascii_str(uni_str))
 
-        asc_str = "".join(ASCII_TO_UNICODE.keys())
+        asc_str = "".join(_asc for _asc, _uco in ASCII_UNICODE)
         assert any(ord(_) < 128 for _ in asc_str)
         assert all(ord(_) < 128 for _ in ascii_str(asc_str))
 
@@ -172,10 +192,10 @@ class TestBaseHelpers:
         uni_str = "äÄßéÉíÍñÑòÒùÙ"
         assert str_ascii(ascii_str(uni_str)) == uni_str
 
-        uni_str = "".join(UNICODE_TO_ASCII.keys())
+        uni_str = "".join(_uco for _asc, _uco in ASCII_UNICODE)
         assert str_ascii(ascii_str(uni_str)) == uni_str
 
-        asc_str = "".join(ASCII_TO_UNICODE.keys())
+        asc_str = "".join(_asc for _asc, _uco in ASCII_UNICODE)
         assert str_ascii(ascii_str(asc_str)) == asc_str
 
     def test_str_ascii_errors(self):
@@ -320,16 +340,18 @@ class TestBaseHelpers:
                 os.remove(tst_fna2)
 
     def test_defuse_maps_integrity(self):
-        assert URI_SEP_CHAR not in UNICODE_TO_ASCII
-        assert len(UNICODE_TO_ASCII) == len(ASCII_TO_UNICODE)   # check for duplicates in the ASCII_UNICODE map
+        assert len(ASCII_TO_UNICODE) == len(ASCII_UNICODE)      # duplicates check in ASCII_UNICODE map
+        assert len(UNICODE_TO_ASCII) == len(ASCII_UNICODE) + 1  # -"-, having also the ord(URI_SEP_UNICODE_CHAR) key
+        assert ord(URI_SEP_UNICODE_CHAR) in UNICODE_TO_ASCII
+        assert UNICODE_TO_ASCII[ord(URI_SEP_UNICODE_CHAR)] == URI_SEP_STR
 
     def test_defuse_maps_not_touching_chars_allowed_as_slug_and_filename(self):
-        assert '-' not in ASCII_TO_UNICODE
-        assert '_' not in ASCII_TO_UNICODE
-        assert '.' not in ASCII_TO_UNICODE
-        assert '~' not in ASCII_TO_UNICODE
+        assert ord('-') not in ASCII_TO_UNICODE
+        assert ord('_') not in ASCII_TO_UNICODE
+        assert ord('.') not in ASCII_TO_UNICODE
+        assert ord('~') not in ASCII_TO_UNICODE
         for char in string.ascii_letters + string.digits:
-            assert char not in ASCII_TO_UNICODE
+            assert ord(char) not in ASCII_TO_UNICODE
 
     def test_dummy_function(self):
         assert dummy_function() is None
@@ -824,6 +846,24 @@ class TestBaseHelpers:
             loaded = parse_dotenv(fp.name)
             assert 'var_nam' not in loaded      # added warning
 
+    def test_parse_dotenv_literal_dict_with_list(self):
+        with tempfile.NamedTemporaryFile(mode="w") as fp:
+            var_val = "{'key': {'sub-key': ['list-item', 'list-item with = char', ]}}"
+            fp.write("var_nam=" + var_val)
+            fp.seek(0)
+            loaded = parse_dotenv(fp.name)
+            assert 'var_nam' in loaded
+            assert loaded['var_nam'] == var_val
+
+    def test_parse_dotenv_multi_line_var_value(self):
+        with tempfile.NamedTemporaryFile(mode="w") as fp:
+            var_val = "{'key': {'sub-key':\\\n    ['list-item',\\\n     'list-item with = char', ]}}"
+            fp.write("var_nam=" + var_val)
+            fp.seek(0)
+            loaded = parse_dotenv(fp.name)
+            assert 'var_nam' in loaded
+            assert loaded['var_nam'] == var_val.replace('\\\n', "")
+
     def test_parse_dotenv_single_in_double_quoted_value(self):
         with tempfile.NamedTemporaryFile(mode="w") as fp:
             fp.write('''var_nam="'var val'"''')
@@ -840,15 +880,6 @@ class TestBaseHelpers:
             assert 'var_nam' in loaded
             assert loaded['var_nam'] == "var val"
 
-    def test_parse_dotenv_literal_dict_with_list(self):
-        with tempfile.NamedTemporaryFile(mode="w") as fp:
-            var_val = "{'key': {'sub-key': ['list-item', 'list-item with = char', ]}}"
-            fp.write("var_nam=" + var_val)
-            fp.seek(0)
-            loaded = parse_dotenv(fp.name)
-            assert 'var_nam' in loaded
-            assert loaded['var_nam'] == var_val
-
     def test_parse_dotenv_literal_dict_with_list_quoted(self):
         with tempfile.NamedTemporaryFile(mode="w") as fp:
             var_val = "{'key': {'sub-key': ['list-item', 'list-item with = char']}}"
@@ -1065,22 +1096,50 @@ class TestBaseHelpers:
         assert to_ascii('ß') == 'ss'
         assert to_ascii('€') == 'Euro'
 
+    @staticmethod
+    def url_failure_httpbin_503_retryer(url: str, timeout: Optional[float] = None) -> tuple[str, str]:
+        retries = 9
+        while True:
+            err_msg = url_failure(url, timeout=timeout)
+            if not err_msg or int(err_msg[:3]) != 503 or retries == 0:
+                break
+            time.sleep(3)
+            retries -= 1
+        return err_msg, f"url_failure({url=}, {timeout=}) httpbin is sometimes unavailable/503. retry later; {err_msg=}"
+
     def test_url_failure(self):
         assert not url_failure("https://gitlab.com/ae-group/ae_base")
 
         assert not url_failure("https://gitlab.com/ae-group/ae_base.git")
 
+        assert not url_failure("https://gitlab.com/ae-group/ae_base", git_repo=True)
+
         assert not url_failure("https://www.google.com")
 
-        assert not url_failure(f"https://httpbin.org/status/200")
-
-    def test_url_failure_errors(self):
-        assert url_failure("")
+        ret, message = self.url_failure_httpbin_503_retryer("https://httpbin.org/status/200")
+        assert not ret, message
 
+    def test_url_failure_authentication_errors(self):
         password, domain, path = "toBeMaskedPassword", "any-not_existing-host_domain.zzz", "any/not/existing/url/path"
         url = f"https://username:{password}@{domain}/{path}"
         err_msg = "raised exception error message"
 
+        ret = url_failure(url, token=password)
+
+        assert ret
+        assert int(ret[:3]) > 0
+        assert password not in ret
+        assert domain in ret
+        assert path in ret
+
+        ret = url_failure(url, username="any user name", password=password)
+
+        assert ret
+        assert int(ret[:3]) > 0
+        assert password not in ret
+        assert domain in ret
+        assert path in ret
+
         ret = url_failure(url)
 
         assert ret
@@ -1089,25 +1148,6 @@ class TestBaseHelpers:
         assert domain in ret
         assert path in ret
 
-        ret = url_failure(url2 := f"https://httpbin.org/status/504")
-
-        assert ret
-        assert int(ret[:3]) == 504
-        assert ret[4:].startswith(mask_url(url2))
-
-        ret = url_failure(f"https://httpbin.org/delay/3", timeout=0.9)
-
-        assert ret
-        assert int(ret[:3]) > 0
-
-        ret = url_failure(f"https://expired.badssl.com")
-
-        assert ret
-        assert int(ret[:3]) > 0
-
-        with pytest.raises(AttributeError):
-            url_failure(cast(str, 123456))
-
         mocked_headers = cast(HTTPMessage, {})
 
         def mock_raise_http_error404(*_args, **_kwargs):
@@ -1200,6 +1240,30 @@ class TestBaseHelpers:
         assert domain in ret
         assert path in ret
 
+    def test_url_failure_ssl_errors(self):
+        ret = url_failure(f"https://expired.badssl.com")
+
+        assert ret
+        assert int(ret[:3]) > 0
+
+    def test_url_failure_timeout_errors(self):
+        ret, message = self.url_failure_httpbin_503_retryer("https://httpbin.org/delay/3", timeout=0.9)
+
+        assert ret, message
+        assert ret[:3] == '997', message
+
+    def test_url_failure_url_errors(self):
+        assert url_failure("")
+
+        ret, message = self.url_failure_httpbin_503_retryer(url2 := "https://httpbin.org/status/504")
+
+        assert ret, message
+        assert ret[:3] == '504', message
+        assert ret[4:].startswith(mask_url(url2)), message
+
+        with pytest.raises(AttributeError):
+            url_failure(cast(str, 123456))
+
     def test_utc_datetime(self):
         dt1 = utc_datetime()
         dt2 = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
